#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn
from .curltoolmixin import CurlToolMixIn


class phpTool(BashToolMixIn, CurlToolMixIn, RuntimeTool):
    """PHP is a popular general-purpose scripting language that is especially suited to web development.

Home: http://php.net/


By default the latest available PHP binary is used for the following OSes:
* Debian & Ubuntu - uses Sury (https://deb.sury.org/) builds 5.6 & 7.x.
* CentOS, RHEL & Oracle Linux - uses SCL 5.6 & 7.x.
* macos uses Homebrew builds 5.6 & 7.x.

You can forbid source builds by setting phpBinOnly to non-empty string.

However, if phpVer is set then we use php-build which make consume a lot of time and
resources due to lack of trusted binary builds.

You can control installed extensions by setting space separated environment:
* phpExtRequire - required extensions to be installed or fail
* phpExtTry - nice to have extensions

The same can be done by setting project-specific .toolTune options as array:
* extRequire = []
* extTry = []
"""
    __slots__ = ()

    def getDeps(self):
        deps = BashToolMixIn.getDeps(self) + CurlToolMixIn.getDeps(self)
        # TODO: need to disable for binary only installs
        deps += ['phpbuild']
        return deps

    def _installTool(self, env):
        ospath = self._ospath
        environ = self._environ
        php_ver = env['phpVer']

        if not env['phpSourceBuild']:
            self._installBinaries(env)
            return

        php_dir = env['phpDir']

        try:
            self._os.makedirs(php_dir)
        except:
            pass

        self._buildDeps(env)

        old_tmpdir = environ.get('TMPDIR', '/tmp')
        environ['TMPDIR'] = ospath.join(php_dir, '..')
        self._executil.callExternal(
            [env['phpbuildBin'], env['phpSourceVer'], env['phpDir']])
        environ['TMPDIR'] = old_tmpdir

    def _installBinaries(self, env):
        detect = self._detect
        phputil = self._phputil
        install = self._install

        ver = env['phpVer']
        php_pkg = phputil.basePackage(ver)

        if detect.isDebian():
            repo = env.get('phpSuryRepo', 'https://packages.sury.org/php')
            gpg = self._callCurl(env, [repo + '/apt.gpg'], binary_output=True)

            install.aptRepo(
                'sury', "deb {0} $codename$ main".format(repo), gpg)
            install.deb(php_pkg)

        elif detect.isUbuntu():
            install.aptRepo('sury', 'ppa:ondrej/php', None)
            install.deb(php_pkg)

        elif detect.isSCLSupported():
            if phputil.isIUSVer(ver):
                install.yumIUS()
                install.yum(php_pkg + '-cli')
            else:
                install.yumSCL()
                install.yum(php_pkg)

        elif detect.isMacOS():
            install.brewTap('homebrew/homebrew-php')
            install.brewUnlink(search='/homebrew\/php\/php[0-9]{2}$/')
            install.brew(php_pkg)

        elif detect.isAlpineLinux():
            install.apkCommunity()
            install.apk(php_pkg)

        elif detect.isSLES():
            self._install.SUSEConnectVerArch('sle-module-web-scripting')
            install.zypper(php_pkg)

        else:
            install.zypper(php_pkg)
            install.yum('{0}-cli'.format(php_pkg))
            install.emerge('dev-lang/php')
            install.pacman(php_pkg)

        env['phpJustInstalled'] = True

    def _updateTool(self, env):
        pass

    def uninstallTool(self, env):
        if env['phpVer'] == self.SYSTEM_VER or env['phpBinOnly']:
            return super(phpTool, self).uninstallTool(env)

        php_dir = env['phpDir']
        self._pathutil.rmTree(php_dir)
        self._have_tool = False

    def envNames(self):
        return ['phpDir', 'phpBin', 'phpVer', 'phpfpmVer', 'phpBinOnly', 'phpSuryRepo',
                'phpExtRequire', 'phpExtTry', 'phpForceBuild', 'phpSourceVer']

    def initEnv(self, env):
        ospath = self._ospath
        os = self._os
        detect = self._detect
        phputil = self._phputil
        environ = self._environ

        #---
        if 'phpfpmVer' in env:
            php_ver = env.setdefault('phpVer', env['phpfpmVer'])

        #---
        phpForceBuild = env.setdefault('phpForceBuild', False)
        phpBinOnly = env.setdefault('phpBinOnly', not phpForceBuild)
        env['phpSourceBuild'] = False

        if phpBinOnly and phpForceBuild:
            self._warn('"phpBinOnly" and "phpForceBuild" do not make sense'
                       ' when set together!')

        php_required_ext = env.get('phpExtRequire', '')
        php_try_ext = env.get('phpExtTry', '')
        environ['phpExtRequire'] = php_required_ext
        environ['phpExtTry'] = php_try_ext
        php_required_ext = php_required_ext.split()
        php_try_ext = php_try_ext.split()

        #---
        php_binaries = phputil.binaryVersions()

        if php_binaries:
            php_latest = php_binaries[-1]
            php_ver = env.setdefault('phpVer', php_latest)
            foundBinary = True

            if php_ver.split('.')[0] == '5' and php_ver != '5.6':
                php_ver = '5.6'
                self._warn('Forcing PHP 5.6 for PHP 5.x requirement')
            elif php_ver == '7' and php_latest.split('.')[0] == '7':
                php_ver = php_latest
            elif php_ver.split('.') > php_latest.split('.'):
                foundBinary = False
                self._warn(
                    'Binary builds are supported only for: {0}'.format(', '.join(php_binaries)))

            env['phpVer'] = php_ver
        else:
            foundBinary = False
            php_ver = env.setdefault('phpVer', self.SYSTEM_VER)

        #---
        if php_ver == self.SYSTEM_VER:
            super(phpTool, self).initEnv(env)
        elif foundBinary and not phpForceBuild:
            if detect.isDebian() or detect.isUbuntu():
                bin_name = 'php' + php_ver
                bin_src = ospath.join('/usr/bin', bin_name)
                self._have_tool = phputil.createBinDir(env, bin_src)

            elif detect.isSCLSupported():
                if phputil.isIUSVer(php_ver):
                    # IUS allows only one version installed as default
                    super(phpTool, self).initEnv(env)
                else:
                    try:
                        ver = php_ver.replace('.', '')
                        sclname = 'rh-php{0}'.format(ver)
                        env_to_set = self._executil.callExternal(
                            ['scl', 'enable', sclname, 'env'], verbose=False)
                        self._pathutil.updateEnvFromOutput(env_to_set)
                        super(phpTool, self).initEnv(env)

                        if self._have_tool:
                            self._have_tool = env['phpBin'].startswith('/opt')
                    except self._ext.subprocess.CalledProcessError:
                        return
                    except OSError:
                        return

            elif detect.isArchLinux():
                if phputil.isArchLatest(php_ver):
                    bin_name = 'php'
                else:
                    bin_name = 'php' + php_ver.replace('.', '')

                bin_src = ospath.join('/usr/bin', bin_name)
                self._have_tool = phputil.createBinDir(env, bin_src)

            elif detect.isAlpineLinux():
                if phputil.isAlpineSplit():
                    bin_name = 'php' + php_ver[0]
                else:
                    bin_name = 'php'

                bin_src = ospath.join('/usr/bin', bin_name)
                self._have_tool = phputil.createBinDir(env, bin_src)

            elif detect.isMacOS():
                brew_prefix = env['brewDir']
                formula = 'php' + php_ver.replace('.', '')
                php_dir = ospath.join(brew_prefix, 'opt', formula, 'bin')

                if ospath.exists(php_dir):
                    self._pathutil.addBinPath(php_dir, True)
                    super(phpTool, self).initEnv(env)

        elif not phpBinOnly:
            def_dir = ospath.join(
                env['phpbuildDir'], 'share', 'php-build', 'definitions')

            if not ospath.exists(def_dir):
                return

            php_ver = env.setdefault('phpSourceVer', php_ver)
            defs = os.listdir(def_dir)
            defs = self._ext.fnmatch.filter(defs, php_ver + '*')

            if not defs:
                self._errorExit('PHP version "{0}" not found'.format(php_ver))

            php_ver = self._versionutil.latest(defs)

            env['phpSourceVer'] = php_ver
            env['phpForceBuild'] = True
            env['phpSourceBuild'] = True

            php_dir = ospath.join(os.environ['HOME'], '.php', php_ver)
            php_dir = env.setdefault('phpDir', php_dir)
            php_bin_dir = ospath.join(php_dir, 'bin')
            php_bin = ospath.join(php_bin_dir, 'php')

            if ospath.exists(php_bin):
                self._have_tool = True
                self._pathutil.addBinPath(php_bin_dir, True)
                env.setdefault('phpBin', php_bin)

        #---
        if self._have_tool:
            if env.get('phpJustInstalled', False):
                self._phputil.installExtensions(env, [
                    'apcu',
                    'curl',
                ], True)

            phputil.installExtensions(env, php_required_ext, False)
            phputil.installExtensions(env, php_try_ext, True)

    def loadConfig(self, config):
        env = config['env']
        phputil = self._phputil

        php_required_ext = self._getTune(config, 'extRequire', [])
        php_try_ext = self._getTune(config, 'extTry', [])

        phputil.installExtensions(env, php_required_ext, False)
        phputil.installExtensions(env, php_try_ext, True)

    def _buildDeps(self, env):
        ospath = self._ospath
        os = self._os
        environ = self._environ

        if self._detect.isSLES():
            self._errorExit(
                'PHP source builds are not supported for SLES yet!')
            return

        self._builddep.require(env, [
            'ssl',
            'mysqlclient',
            'postgresql',
        ])
        # APT
        #---
        self._install.deb([
            'build-essential',
            'bison',
            'automake',
            'autoconf',
            'libtool',
            're2c',
            'libcurl4-openssl-dev',
            'libtidy-dev',
            'libpng-dev',
            'libmcrypt-dev',
            'libjpeg-dev',
            'libreadline-dev',
            'libbz2-dev',
            'libc-client-dev',
            'libdb-dev',
            'libedit-dev',
            'libenchant-dev',
            'libevent-dev',
            'libexpat1-dev',
            'libfreetype6-dev',
            'libgcrypt11-dev',
            'libgd2-dev',
            'libglib2.0-dev',
            'libgmp3-dev',
            'libicu-dev',
            'libjpeg-dev',
            'libkrb5-dev',
            'libldap2-dev',
            'libmagic-dev',
            'libmhash-dev',
            'libonig-dev',
            'libpam0g-dev',
            'libpcre3-dev',
            'libpng-dev',
            'libpspell-dev',
            'libqdbm-dev',
            'librecode-dev',
            'libsasl2-dev',
            'libsnmp-dev',
            'libsqlite3-dev',
            'libwrap0-dev',
            'libxmltok1-dev',
            'libxml2-dev',
            'libvpx-dev',
            'libxslt1-dev',
            'unixodbc-dev',
            'zlib1g-dev',
        ])

        # Extra repo before the rest
        #---
        self._install.yumEPEL()

        self._install.rpm([
            'binutils',
            'patch',
            'git',
            'gcc',
            'gcc-c++',
            'make',
            'autoconf',
            'automake',
            'libtool',
            'bison',
            're2c',
            'glibc-devel',
            'libxml2-devel',
            'pkgconfig',
            'curl-devel',
            'libpng-devel',
            'libjpeg-devel',
            'libXpm-devel',
            'freetype-devel',
            'gmp-devel',
            'libmcrypt-devel',
            'aspell-devel',
            'recode-devel',
            'libicu-devel',
            'oniguruma-devel',
            'libtidy-devel',
            'libxslt-devel',
            'readline-devel',
            'zlib-devel',
            'pcre-devel',
        ])

        self._install.yum('bzip2-devel')
        self._install.zypper('libbz2-devel')

        self._install.emergeDepsOnly(['dev-lang/php'])
        self._install.pacman([
            'patch',
            'git',
            'gcc',
            'make',
            'autoconf',
            'automake',
            'libtool',
            'bison',
            're2c',
            'glibc',
            'libxml2',
            'curl',
            'libpng',
            'libjpeg',
            'libxpm',
            'freetype2',
            'gmp',
            'libmcrypt',
            'aspell',
            'recode',
            'icu',
            'oniguruma',
            'tidy',
            'libxslt',
            'readline',
            'zlib',
            'pcre',
        ])

        #---
        systemctl = self._pathutil.which('systemctl')

        if systemctl:
            self._install.deb(['libsystemd-dev'])
            self._install.rpm(['systemd-devel'])
            with_systemd = ' --with-fpm-systemd'
        else:
            with_systemd = ' --without-fpm-systemd'

        multiarch = None
        dpkgarch = self._pathutil.which('dpkg-architecture')

        if dpkgarch:
            multiarch = self._executil.callExternal(
                [dpkgarch, '-qDEB_HOST_MULTIARCH']).strip()

        if multiarch:
            if ospath.exists(ospath.join('/usr/include', multiarch, 'curl')):
                curl_dir = ospath.join(env['phpDir'], '..', 'curl')

                try:
                    os.mkdir(curl_dir)
                    os.symlink(ospath.join('/usr/include', multiarch),
                               ospath.join(curl_dir, 'include'))
                    os.symlink(ospath.join('/usr/lib', multiarch),
                               ospath.join(curl_dir, 'lib'))
                except Exception as e:
                    # print(e)
                    pass
            else:
                curl_dir = '/usr/include'

            with_libdir = ' --with-libdir={0} --with-curl={1}'.format(
                ospath.join('lib', multiarch),
                curl_dir,
            )
        else:
            with_libdir = ''
        #---
        from ..details.resourcealgo import ResourceAlgo
        cpu_count = ResourceAlgo().cpuLimit({})

        environ['PHP_BUILD_EXTRA_MAKE_ARGUMENTS'] = '-j{0}'.format(
            cpu_count)

        environ['PHP_BUILD_CONFIGURE_OPTS'] = ' \
            --disable-debug \
            --with-regex=php \
            --enable-calendar \
            --enable-sysvsem \
            --enable-sysvshm \
            --enable-sysvmsg \
            --enable-bcmath \
            --disable-cgi \
            --disable-phpdbg \
            --enable-fpm \
            --with-bz2 \
            --enable-ctype \
            --without-db4 \
            --without-qdbm \
            --without-gdbm \
            --with-iconv \
            --enable-exif \
            --enable-ftp \
            --with-gettext \
            --enable-mbstring \
            --with-onig=/usr \
            --with-pcre-regex=/usr \
            --enable-shmop \
            --enable-sockets \
            --enable-wddx \
            --with-libxml-dir=/usr \
            --with-zlib \
            --with-kerberos=/usr \
            --with-openssl=/usr \
            --enable-soap \
            --enable-zip \
            --with-mhash=yes \
            --with-system-tzdata \
            ' + with_systemd + with_libdir

    def tuneDefaults(self, env):
        return {
            'minMemory': '8M',
            'socketType': 'none',
            'scalable': False,
            'reloadable': False,
            'multiCore': False,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .curltoolmixin import CurlToolMixIn


class jfrogTool(CurlToolMixIn, RunEnvTool):
    """JFrog: Command Line Interface for Artifactory and Bintray

Home: https://www.jfrog.com/confluence/display/CLI/JFrog+CLI
"""
    __slots__ = ()

    def _installTool(self, env):
        ospath = self._ospath
        os = self._os

        if self._detect.isMacOS():
            self._install.brew('jfrog-cli-go')
            return

        dst_dir = env['jfrogDir']
        get_url = env['jfrogGet']
        jfrog_bin = ospath.join(dst_dir, 'jfrog')

        if not ospath.exists(dst_dir):
            os.makedirs(dst_dir)

        self._callCurl(env, [get_url, '-o', jfrog_bin])
        stat = self._ext.stat
        os.chmod(jfrog_bin, stat.S_IRWXU | stat.S_IRGRP |
                 stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

    def _updateTool(self, env):
        if self._detect.isMacOS():
            return

        self.uninstallTool(env)
        self._installTool(env)

    def uninstallTool(self, env):
        if self._detect.isMacOS():
            return

        jfrog_bin = env['jfrogBin']
        if self._ospath.exists(jfrog_bin):
            self._os.remove(jfrog_bin)
        self._have_tool = False

    def envNames(self):
        return ['jfrogDir', 'jfrogBin', 'jfrogGet']

    def initEnv(self, env):
        bin_dir = env.setdefault('jfrogDir', env['binDir'])

        pkg = None
        url_base = 'https://api.bintray.com/content/jfrog/jfrog-cli-go/$latest'

        detect = self._detect

        if detect.isMacOS():
            pass
        elif detect.isAMD64():
            pkg = 'jfrog-cli-linux-amd64'
        else:
            pkg = 'jfrog-cli-linux-386'

        if pkg:
            env.setdefault(
                'jfrogGet',
                'https://api.bintray.com/content/jfrog/jfrog-cli-go/$latest/{0}/jfrog?bt_package={0}'.format(
                    pkg)
            )

        self._pathutil.addBinPath(bin_dir)

        super(jfrogTool, self).initEnv(env)

        if self._have_tool:
            env['jfrogDir'] = self._ospath.dirname(env['jfrogBin'])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..subtool import SubTool


class PipToolMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['pip', 'virtualenv']

    def _pipName(self):
        return self._name

    def _installTool(self, env):
        self._builddep.require(env, 'python')

        self._executil.callExternal(
            [env['pipBin'], 'install', '-q', self._pipName()])

    def installTool(self, env):
        if not self._have_tool:
            self._installTool(env)
            self.initEnv(env)

            if not self._have_tool:
                self._errorExit('Failed to install "{0}"'.format(self._name))

    def _updateTool(self, env):
        self._executil.callExternal([env['pipBin'], 'install', '-q',
                                     '--upgrade', self._pipName()])

    def uninstallTool(self, env):
        self._executil.callExternal([env['pipBin'], 'uninstall',
                                     '--yes', '-q', self._pipName()])
        self._have_tool = False

    def initEnv(self, env, bin_name=None):
        name = self._name
        bin_env = name + 'Bin'

        if not env.get(bin_env, None):
            if bin_name is None:
                bin_name = name

            ospath = self._ospath
            tool_path = ospath.join(env['virtualenvDir'], 'bin', bin_name)

            if ospath.exists(tool_path):
                env[bin_env] = tool_path
                self._have_tool = True
        else:
            self._have_tool = True

    def _requirePythonDev(self, env):
        if int(env['pythonVer'].split('.')[0]) == 3:
            self._install.deb(['python3-dev'])
            self._install.zypper(['python3-devel'])
            self._install.yumEPEL()
            self._install.yum(['python34-devel'])
            self._install.pacman(['python'])
            self._install.apk('python3-dev')
        else:
            self._install.deb(['python-dev'])
            self._install.zypper(['python-devel'])
            self._install.yumEPEL()
            self._install.yum(['python-devel'])
            self._install.pacman(['python2'])
            self._install.apk('python2-dev')

        self._install.emergeDepsOnly(['dev-lang/python'])

    def _requirePip(self, env, package):
        self._executil.callExternal([env['pipBin'], 'install', '-q', package])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from __future__ import print_function

from collections import OrderedDict

from .mixins.data import DataSlots
from .mixins.tool import ToolMixIn
from .mixins.lock import LockMixIn
from .mixins.config import ConfigMixIn
from .mixins.deploy import DeployMixIn
from .mixins.service import ServiceMixIn
from .mixins.log import LogMixIn


from .coloring import Coloring

from .vcstool import VcsTool
from .rmstool import RmsTool
from .buildtool import BuildTool
from .testtool import TestTool
from .migrationtool import MigrationTool
from .runtimetool import RuntimeTool


__all__ = ['CIDTool']


def cid_action(f):
    def custom_f(self, *args, **kwargs):
        self._processWcDir()
        config = self._config

        try:
            fn = f.func_name
        except AttributeError:
            fn = f.__name__

        actions = config.get('actions', {})

        if fn in actions:
            self._call_actions(fn, actions, args, f)
        else:
            f(self, *args, **kwargs)
    return custom_f


#=============================================================================
class CIDTool(LogMixIn, ConfigMixIn, LockMixIn, ServiceMixIn, DeployMixIn, ToolMixIn, DataSlots):
    __slots__ = ()

    def __init__(self, overrides):
        super(CIDTool, self).__init__()
        self._overrides = overrides
        self._initConfig(True)

    def _call_actions(self, name, actions, args, orig_action=False, filt_args=None):
        act = actions[name]
        act = self._configutil.listify(act)

        if filt_args is None:
            # replace possible None with empty strings
            filt_args = [x or '' for x in args]

            # trim only ending
            while len(filt_args) and not len(filt_args[-1]):
                filt_args.pop()

        for cmd in act:
            if cmd.startswith('@cid'):
                cmd = self._ext.shlex.split(cmd)
                self._executil.callExternal([self._sys.executable, '-mfutoin.cid'] + cmd[1:] + filt_args,
                                            user_interaction=True)
            elif cmd == '@default':
                if orig_action:
                    orig_action(self, *args)
                else:
                    self._errorExit(
                        '@default is not allowed for "[0}"'.format(name))
            elif cmd in actions:
                self._call_actions(cmd, actions, args, False, filt_args)
            else:
                if filt_args:
                    cmd = '{0} {1}'.format(
                        cmd, self._ext.subprocess.list2cmdline(filt_args))
                self._executil.callExternal(
                    ['sh', '-c', cmd], user_interaction=True)

    @cid_action
    def tag(self, branch, next_version=None):
        # implciit in @cid_action
        # self._processWcDir()

        mode = 'patch'

        if next_version in ['patch', 'minor', 'major']:
            mode = next_version
            next_version = None

        if next_version and not self._ext.re.match('^[0-9]+\.[0-9]+\.[0-9]+$', next_version):
            self._errorExit('Valid version format: x.y.z')

        config = self._config
        vcstool = self._getVcsTool()
        vcsrepo = config['vcsRepo']

        #---
        self._info(
            'Getting source branch {0} from {1}'.format(branch, vcsrepo))
        vcstool.vcsCheckout(config, branch)
        self._initConfig()
        config = self._config

        if 'name' not in config:
            self._errorExit('Failed to detect project "name"')

        # Set new version
        if next_version is None:
            if 'version' in config:
                next_version = config['version']
            else:
                self._errorExit('current project version is unknown')

            next_version = next_version.split('.')
            next_version += ['0'] * 3
            del next_version[3:]

            if mode == 'patch':
                next_version[2] = str(int(next_version[2]) + 1)
            elif mode == 'minor':
                next_version[1] = str(int(next_version[1]) + 1)
                next_version[2] = '0'
            elif mode == 'major':
                next_version[0] = str(int(next_version[0]) + 1)
                next_version[1] = '0'
                next_version[2] = '0'

            next_version = '.'.join(next_version)

        config['version'] = next_version

        #---
        self._info('Updating files for release')
        to_commit = []
        self._forEachTool(
            lambda config, t: to_commit.extend(
                t.updateProjectConfig(config, {'version': next_version})
            )
        )

        #---
        if to_commit:
            self._info('Committing updated files')
            message = "Updated for release %s %s" % (
                config['name'], config['version'])
            vcstool.vcsCommit(config, message, to_commit)
        else:
            self._info('Nothing to commit')

        #---
        tag = "v%s" % next_version
        self._info('Creating a tag {0}'.format(tag))
        message = "Release %s %s" % (config['name'], config['version'])
        vcstool.vcsTag(config, tag, message)

        #---
        self._info('Pushing changes to {0}'.format(vcsrepo))
        vcstool.vcsPush(config, [branch, tag])

    @cid_action
    def prepare(self, vcs_ref=None):
        # implciit in @cid_action
        # self._processWcDir()

        config = self._config

        # make a clean checkout
        if vcs_ref:
            vcstool = self._getVcsTool()

            self._info('Getting source ref {0} from {1}'.format(
                vcs_ref, config['vcsRepo']))
            vcstool.vcsCheckout(config, vcs_ref)
            self._initConfig()

        #--
        self._info('Running "prepare" in tools')
        self._forEachTool(
            lambda config, t: t.onPrepare(config),
            base=BuildTool
        )

    @cid_action
    def build(self):
        # implciit in @cid_action
        # self._processWcDir()

        self._info('Running "build" in tools')
        self._forEachTool(
            lambda config, t: t.onBuild(config),
            base=BuildTool
        )

    @cid_action
    def package(self):
        # implciit in @cid_action
        # self._processWcDir()

        #---
        self._info('Running "package" in tools')
        self._forEachTool(
            lambda config, t: t.onPackage(config),
            base=BuildTool
        )

        #---
        os = self._os
        ospath = self._ospath
        pathutil = self._pathutil
        fnmatch = self._ext.fnmatch
        config = self._config
        package_files = config.get('packageFiles', None)

        if package_files is not None:
            self._info(
                'Found binary artifacts from tools: {0}'.format(package_files))
            self._lastPackages = package_files
            return

        #---
        try:
            package_content = config['package']
        except KeyError:
            package_content = set(os.listdir('.'))
            # TODO: make it more extensible
            package_content -= set(
                fnmatch.filter(package_content, '.git*') +
                fnmatch.filter(package_content, '.hg*') +
                ['.svn']
            )
            package_content = list(package_content)

        if type(package_content) != list:
            package_content = [package_content]

        package_content.sort()

        self._info('Generating package from {0}'.format(package_content))

        #---
        if config.get('packageChecksums', True):
            self._info('Generating checksums')
            checksums_file = '.package.checksums'

            checksums = []
            cs_files = []

            for pkg_item in sorted(package_content):
                if ospath.isfile(pkg_item):
                    cs_files.append(pkg_item)

                if ospath.isdir(pkg_item):
                    for (path, dirs, files) in os.walk(pkg_item):
                        for f in sorted(files):
                            f = pathutil.safeJoin(path, f)
                            cs_files.append(f)

            hashlib = self._ext.hashlib

            for cf in cs_files:
                hasher = hashlib.sha512()

                with open(cf, 'rb') as f_in:
                    for chunk in iter(lambda: f_in.read(65536), b""):
                        hasher.update(chunk)

                checksums.append("{0}  {1}".format(hasher.hexdigest(), cf))

            checksums.append('')

            with open(checksums_file, 'w') as f_out:
                f_out.write("\n".join(checksums))

            try:
                package_content.index('.')
            except ValueError:
                package_content.append(checksums_file)

        #---
        buildTimestamp = self._ext.datetime.datetime.utcnow().strftime('%Y%m%d%H%M%S')
        name = config.get('name', 'UNKNOWN').split('/')[-1]
        version = config.get('version', 'UNKNOWN')
        vcs_ref = config.get('vcsRef', None)

        # Note: unless run in clean ci_build process,
        # all builds must be treated as snapshots/CI builds
        if vcs_ref == 'v' + version:
            package_name = '{0}-{1}-{2}'.format(
                name, version, buildTimestamp)
        else:
            vcs_ref = 'UNKNOWN'

            if config.get('vcs', None):
                try:
                    vcstool = self._getVcsTool()
                    vcs_ref = vcstool.vcsGetRevision(config)
                except self._ext.subprocess.CalledProcessError as e:
                    if config.get('vcsRepo', None):
                        raise e

            package_name = '{0}-CI-{1}-{2}-{3}'.format(
                name, version, buildTimestamp, vcs_ref)

        if 'target' in config:
            package_name += '-{0}'.format(config['target'])

        package_file = package_name + '.txz'
        self._info('Creating package {0}'.format(package_file))

        tar_tool = self._getTarTool('xz')
        tar_args = ['cJf', package_file,
                    '--transform=s,^,{0}/,'.format(package_name),
                    '--exclude=' + package_file,
                    '--exclude=.git*',
                    '--exclude=.hg*',
                    '--exclude=.svn'] + package_content
        tar_tool.onExec(config['env'], tar_args, False)
        # note, no --exclude-vcs
        self._lastPackages = [package_file]

    @cid_action
    def check(self):
        # implciit in @cid_action
        # self._processWcDir()

        self._info('Running "check" in tools')
        self._forEachTool(
            lambda config, t: t.onCheck(config),
            allow_failure=self._config.get('permissiveChecks', False),
            base=TestTool
        )

    @cid_action
    def promote(self, rms_pool, packages):
        # implciit in @cid_action
        # self._processWcDir()

        config = self._config
        rmstool = self._getRmsTool()

        pools = rms_pool.split(':', 1)

        if len(pools) == 2:
            self._info('Promoting from {0} to {1} pool: {2}'.format(
                pools[0], pools[1], ', '.join(packages)))
            packages = rmstool.rmsProcessChecksums(config, pools[0], packages)
            rmstool.rmsPromote(config, pools[0], pools[1], packages)
        else:
            self._info('Promoting to {0} pool: {1}'.format(
                rms_pool, ', '.join(packages)))
            rmstool.rmsUpload(config, rms_pool, packages)

    @cid_action
    def migrate(self):
        # implciit in @cid_action
        # self._processWcDir()

        self._info('Running "migrate" in tools')
        self._forEachTool(
            lambda config, t: t.onMigrate(config),
            base=MigrationTool
        )

    def deploy(self, mode, p1=None, p2=None):
        self._processDeployDir()

        self._deployLock()

        try:
            if mode == 'rms':
                self._rms_deploy(p1, p2)
            elif mode == 'vcsref':
                self._vcsref_deploy(p1)
            elif mode == 'vcstag':
                self._vcstag_deploy(p1)
            elif mode == 'setup':
                self._deploy_setup()
            else:
                self._errorExit('Not supported deploy mode: ' + mode)
        finally:
            self._deployUnlock()

    def deploy_set(self, action, *args):
        self._processDeployDir()

        member = '_deploy_set_{0}'.format(action)
        getattr(self, member)(*args)

        self._deployLock()

        try:
            self._writeDeployConfig()
        finally:
            self._deployUnlock()

    def deploy_reset(self, action):
        self._processDeployDir()

        if action:
            actions = [action]
        else:
            actions = [
                'tools',
                'tooltune',
                'action',
                'persistent',
                'writable',
                'entrypoint',
                'env',
                'webcfg',
                #'webmount',
            ]

        for action in actions:
            member = '_deploy_reset_{0}'.format(action)
            getattr(self, member)()

        self._deployLock()

        try:
            self._writeDeployConfig()
        finally:
            self._deployUnlock()

    def run(self, command, args):
        self._processWcDir()

        os = self._os
        config = self._config
        entry_points = config.get('entryPoints', {})
        actions = config.get('actions', {})
        args = args or []

        if command:
            if command in entry_points:
                ep = entry_points[command]

                # Do stuff required to get the tool ready
                tool = ep['tool']
                self._overrides['tool'] = tool
                self._initConfig()

                #
                config = self._config
                t = self._getTool(tool)

                if isinstance(t, RuntimeTool):
                    svc = self._ext.copy.deepcopy(ep)
                    svc['name'] = command
                    svc['instanceId'] = 0
                    svc.setdefault('tune', {})

                    runtime_dir = self._ext.tempfile.mkdtemp(
                        prefix='futoin-cid-run')

                    t.onPreConfigure(config, runtime_dir, svc, svc['tune'])
                    t.onRun(config, svc, args)
                else:
                    self._errorExit(
                        'Tool "{0}" for "{1}" does not support "run" command'.format(tool, command))

            elif command in actions:
                self._call_actions(command, actions, args)
            else:
                self._errorExit(
                    'Unknown "{0}" action or entry point'.format(command))
        else:
            for cmd in entry_points:
                pid = os.fork()

                if not pid:
                    os.dup2(os.open(os.devnull, os.O_RDONLY), 0)
                    self.run(cmd, None)

            i = len(entry_points)

            while i > 0:
                os.wait()
                i -= 1

    def ci_build(self, vcs_ref, rms_pool):
        self._initConfig()

        os = self._os
        config = self._config
        wcDir = config['wcDir']

        if self._ospath.exists(wcDir) and wcDir != os.getcwd():
            try:
                dst = '{0}.bak{1}'.format(wcDir, int(self._ext.time.time()))
                self._infoLabel('Renaming: ', '{0} to {1}'.format(wcDir, dst))
                os.rename(wcDir, dst)
            except OSError:
                self._pathutil.rmTree(wcDir)

        # Make sure to keep VCS info when switch to another location
        # for checkout.
        #---
        if 'vcs' in config:
            self._getVcsTool()

            for cv in ['vcs', 'vcsRepo']:
                self._overrides[cv] = config[cv]
        #---

        self._processWcDir()

        self._lastPackages = None
        self.prepare(vcs_ref)
        self.build()
        self.package()
        self.check()

        if rms_pool and self._lastPackages:
            self.promote(rms_pool, self._lastPackages)

    def tool_exec(self, tool, args):
        self._processWcDir()
        t = self._getTool(tool)
        t.onExec(self._env, args)

    def tool_envexec(self, tool, command):
        self._processWcDir()
        self._executil.callInteractive(command, search_path=True)

    def tool_install(self, tool):
        self._processWcDir()

        if self._detect.isDisabledToolsSetup(self._env):
            self._errorExit(
                'environment requires external installation of tools')

        if tool:
            tools = [tool]
        else:
            tools = self._config['toolOrder']

        env = self._env

        self._globalLock()

        for tool in tools:
            t = self._getTool(tool)
            t.requireInstalled(env)

        self._globalUnlock()

    def tool_uninstall(self, tool):
        self._processWcDir()

        if self._detect.isDisabledToolsSetup(self._env):
            self._errorExit(
                'environment requires external management of tools')

        if tool:
            tools = [tool]
        else:
            tools = reversed(self._config['toolOrder'])

        env = self._env

        self._globalLock()

        for tool in tools:
            t = self._getTool(tool)
            if t.isInstalled(env):
                t.uninstallTool(env)

        self._globalUnlock()

    def tool_update(self, tool):
        self._processWcDir()

        if self._detect.isDisabledToolsSetup(self._env):
            self._errorExit(
                'environment requires external management of tools')

        if tool:
            tools = [tool]
        else:
            tools = self._config['toolOrder']

        env = self._env

        self._globalLock()

        for tool in tools:
            t = self._getTool(tool)
            t.updateTool(env)

        self._globalUnlock()

    def tool_test(self, tool):
        self._processWcDir()

        if tool:
            tools = [tool]
        else:
            tools = self._config['toolOrder']

        env = self._env

        for tool in tools:
            t = self._getTool(tool)

            if not t.isInstalled(env):
                ver = env.get(tool + 'Ver', None)

                if ver:
                    self._errorExit(
                        "Tool '{0}' version '{1}' is missing".format(tool, ver))
                else:
                    self._errorExit("Tool '{0}' is missing".format(tool))

    def tool_env(self, tool):
        self.tool_test(tool)

        if tool:
            tools = [tool]
        else:
            tools = self._config['toolOrder']

        res = dict(self._environ)
        env = self._env

        # remove unchanged vars
        for k, v in self._origEnv().items():
            if res.get(k, None) == v:
                del res[k]

        for tool in tools:
            self._getTool(tool).exportEnv(env, res)

        for k, v in sorted(res.items()):
            if type(v) == type(''):
                v = v.replace("'", "\\'").replace('\\', '\\\\')
            elif not v:
                v = ''
            print("{0}='{1}'".format(k, v))

    def _tool_cmd(self, tool, base, method):
        self._processWcDir()

        config = self._config
        t = self._getTool(tool)

        if isinstance(t, base):
            t.loadConfig(config)  # see self._initTools()
            getattr(t, method)(config)
        else:
            self._errorExit(
                '{0} tool does not support {1}'.format(tool, method))

    def tool_prepare(self, tool):
        self._tool_cmd(tool, BuildTool, 'onPrepare')

    def tool_build(self, tool):
        self._tool_cmd(tool, BuildTool, 'onBuild')

    def tool_check(self, tool):
        self._tool_cmd(tool, TestTool, 'onCheck')

    def tool_package(self, tool):
        self._tool_cmd(tool, BuildTool, 'onPackage')

        package_files = self._config.get('packageFiles', None)

        if package_files is not None:
            self._info('Package files: {0}'.format(package_files))

    def tool_migrate(self, tool):
        self._tool_cmd(tool, MigrationTool, 'onMigrate')

    def tool_list(self):
        print("List of tools supported by CID:")
        for k in sorted(self._getKnownTools()):
            t = self._getTool(k)
            doc = t.__doc__.strip() or Coloring.warn('!! Missing documentation.')
            doc = doc.split("\n")[0]
            print(Coloring.infoLabel("  * " + k + ': ') + Coloring.info(doc))
        print('End.')

    def tool_describe(self, tool):
        t = self._getTool(tool)

        print(Coloring.infoLabel('* Tool: ') + Coloring.warn(tool))

        auto_detect = t.autoDetectFiles()
        if auto_detect:
            auto_detect = self._configutil.listify(auto_detect)
            print(Coloring.infoLabel('* Auto-detection (files): ') +
                  ', '.join(auto_detect))

        if isinstance(t, VcsTool):
            print(Coloring.infoLabel('* Auto-detected, if set as VCS '))

        if isinstance(t, RmsTool):
            print(Coloring.infoLabel('* Auto-detected, if set as RMS '))

        env_vars = t.envNames()
        if env_vars:
            print(Coloring.infoLabel(
                '* Environment variables: ') + ', '.join(env_vars))

        deps = t.getDeps()
        if deps:
            print(Coloring.infoLabel('* Dependencies: ') + ', '.join(deps))

        rdeps = t.getPostDeps()
        if rdeps:
            print(Coloring.infoLabel('* Reverse dependencies: ') + ', '.join(rdeps))

        order = t.getOrder()
        if order:
            print(Coloring.infoLabel('* Order shift: ') + str(order))

        print()
        doc = t.__doc__.strip() or Coloring.warn('!! Missing documentation.')
        print(Coloring.info(doc))
        print()

    def tool_detect(self):
        self._processWcDir()

        config = self._config
        env = config['env']

        for t in config['toolOrder']:
            ver = env.get(t + 'Ver', None)

            if ver:
                print("{0}={1}".format(t, ver))
            else:
                print(t)

    def init_project(self, project_name):
        self._processWcDir()
        ospath = self._ospath

        if ospath.exists(self._FUTOIN_JSON):
            self._errorExit('futoin.json already exists in project root')

        config = self._config
        new_config = self._exportConfig(config)

        if project_name:
            new_config['name'] = project_name
        elif 'name' not in config:
            new_config['name'] = ospath.basename(config['wcDir'])

        self._pathutil.writeJSONConfig(self._FUTOIN_JSON, new_config)

    def vcs_checkout(self, vcs_ref):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Getting source ref {0} from {1}'.format(
            vcs_ref, config['vcsRepo']))
        vcstool.vcsCheckout(config, vcs_ref)

    def vcs_commit(self, msg, files):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        if files:
            self._info('Committing: ' + ', '.join(files))
        else:
            self._info('Committing updated files')
        vcstool.vcsCommit(config, msg, files)

        self._info('Pushing changes to {0}'.format(config['vcsRepo']))
        vcstool.vcsPush(config, None)

    def vcs_branch(self, vcs_ref):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Creating new branch {0} in {1}'.format(
            vcs_ref, config['vcsRepo']))
        vcstool.vcsBranch(config, vcs_ref)

    def vcs_merge(self, vcs_ref, cleanup=True):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Merging branch {0} from {1}'.format(
            vcs_ref, config['vcsRepo']))
        vcstool.vcsMerge(config, vcs_ref, cleanup)

    def vcs_delete(self, vcs_ref, vcs_cache_dir):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Deleting branch {0} from {1}'.format(
            vcs_ref, config['vcsRepo']))
        vcstool.vcsDelete(config, vcs_cache_dir, vcs_ref)

    def vcs_export(self, vcs_ref, dst_path, vcs_cache_dir):
        os = self._os

        if self._ospath.exists(dst_path):
            if os.listdir(dst_path):
                self._errorExit(
                    'Destination directory {0} exists and is not empty'.format(dst_path))
        else:
            os.makedirs(dst_path)

        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Export ref {0} from {1}'.format(
            vcs_ref, config['vcsRepo']))
        vcstool.vcsExport(config, vcs_cache_dir, vcs_ref, dst_path)

    def vcs_tags(self, tag_hint, vcs_cache_dir):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        tag_list = vcstool.vcsListTags(config, vcs_cache_dir, tag_hint)

        if tag_hint:
            tag_list = self._ext.fnmatch.filter(tag_list, tag_hint)

        self._versionutil.sort(tag_list)

        print("\n".join(tag_list))

    def vcs_branches(self, branch_hint, vcs_cache_dir):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        branch_list = vcstool.vcsListBranches(
            config, vcs_cache_dir, branch_hint)

        if branch_hint:
            branch_list = self._ext.fnmatch.filter(branch_list, branch_hint)

        self._versionutil.sort(branch_list)

        print("\n".join(branch_list))

    def vcs_reset(self):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Reverting all local changes')
        vcstool.vcsRevert(config)

    def vcs_ismerged(self, vcs_ref):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        if vcstool.vcsIsMerged(config, vcs_ref):
            self._info('Branch {0} is merged'.format(vcs_ref))
        else:
            self._info('Branch {0} is NOT merged'.format(vcs_ref))
            self._sys.exit(1)

    def vcs_clean(self):
        self._processWcDir()

        config = self._config
        vcstool = self._getVcsTool()

        self._info('Removing all unversioned items.')
        vcstool.vcsClean(config)

    def rms_list(self, rms_pool, package_pattern):
        self._processWcDir()

        config = self._config
        rmstool = self._getRmsTool()

        package_list = rmstool.rmsGetList(config, rms_pool, package_pattern)

        if package_pattern:
            package_list = self._ext.fnmatch.filter(
                package_list, package_pattern)

        self._versionutil.sort(package_list)

        print("\n".join(package_list))

    def rms_retrieve(self, rms_pool, package_list):
        self._processWcDir()

        ospath = self._ospath
        config = self._config
        rmstool = self._getRmsTool()

        self._info('Retrieving packages from RMS...')

        for p in package_list:
            p = ospath.basename(p).split('@', 1)[0]

            if ospath.exists(p):
                self._errorExit('File already exists: {0}'.format(p))

        package_list = rmstool.rmsProcessChecksums(
            config, rms_pool, package_list)
        rmstool.rmsRetrieve(config, rms_pool, package_list)

    def rms_pool_create(self, rms_pool):
        self._processWcDir()

        config = self._config
        rmstool = self._getRmsTool()

        self._info('Creating RMS pool...')

        rmstool.rmsPoolCreate(config, rms_pool)

    def rms_pool_list(self):
        self._processWcDir()

        config = self._config
        rmstool = self._getRmsTool()

        pool_list = rmstool.rmsPoolList(config)

        pool_list = sorted(pool_list)

        print("\n".join(pool_list))

    def service_master(self):
        self._processDeployDir()
        self._serviceAdapt()
        self._serviceMaster()

    def service_list(self):
        self._processDeployDir()
        self._serviceAdapt()
        self._serviceListPrint()

    def service_exec(self, entry_point, instance_id):
        self._processDeployDir()

        config = self._config
        svc = self._serviceCommon(entry_point, instance_id)

        tool = svc['tool']
        t = self._getTool(tool)

        if isinstance(t, RuntimeTool):
            self._os.chdir(config['wcDir'])
            t.onRun(config, svc, [])
        else:
            self._errorExit(
                'Tool "{0}" for "{1}" does not support "service exec" command'.format(tool, entry_point))

    def service_stop(self, entry_point, instance_id, pid):
        self._processDeployDir()

        svc = self._serviceCommon(entry_point, instance_id)

        tool = svc['tool']
        t = self._getTool(tool)

        if isinstance(t, RuntimeTool):
            self._serviceStop(svc, t, int(pid))
        else:
            self._errorExit(
                'Tool "{0}" for "{1}" does not support "service stop" command'.format(tool, entry_point))

    def service_reload(self, entry_point, instance_id, pid):
        self._processDeployDir()

        config = self._config
        svc = self._serviceCommon(entry_point, instance_id)

        tool = svc['tool']
        t = self._getTool(tool)

        if isinstance(t, RuntimeTool):
            t.onReload(config, int(pid), svc['tune'])
        else:
            self._errorExit(
                'Tool "{0}" for "{1}" does not support "service reload" command'.format(tool, entry_point))

    def devserve(self):
        self._devserve_mode = True

        ospath = self._ospath
        pathutil = self._pathutil
        os = self._os

        deploy_dir = self._ext.tempfile.mkdtemp(prefix='futoin-cid-devserve')

        try:
            wc_dir = ospath.realpath(self._overrides['wcDir'])
            deploy_dir = ospath.realpath(deploy_dir)
            os.symlink(wc_dir, pathutil.safeJoin(deploy_dir, 'current'))

            self._overrides['devWcDir'] = wc_dir
            self._overrides['deployDir'] = deploy_dir

            self._deployLock()
            self._deployUnlock()

            self._processDeployDir()
            self._env['type'] = 'dev'
            self._serviceAdapt()
            self._serviceListPrint()
            self._serviceMaster()
            pathutil.rmTree(deploy_dir)
            deploy_dir = None
        finally:
            if deploy_dir:
                self._warn(
                    'Left "{0}" for inspection of error'.format(deploy_dir))

    def sudoers(self, entity, skip_key_mgmt):
        ospath = self._ospath
        detect = self._detect

        if not entity:
            entity = self._ext.pwd.getpwuid(self._os.geteuid())[0]

        lines = ['']
        commands = []

        if detect.isDebian() or detect.isUbuntu():
            lines += [
                '# env whitelist',
                'Defaults  env_keep += "DEBIAN_FRONTEND"',
            ]
            commands += [
                '# package installation only',
                '/usr/bin/apt-get install *',
                '/usr/bin/apt-get update',
                '/usr/bin/apt-add-repository *',
                '/usr/bin/dpkg -i *',
            ]

            if not skip_key_mgmt:
                commands += [
                    '',
                    '# signing key setup',
                    '/usr/bin/apt-key add *',
                ]

        elif detect.isFedora():
            commands += [
                '# package installation only',
                '/usr/bin/dnf install *',
                '/usr/bin/dnf config-manager --add-repo *',
                '/usr/bin/dnf config-manager --enable *',
            ]

            if not skip_key_mgmt:
                commands += [
                    '',
                    '# signing key setup',
                    '/usr/bin/rpm --import *',
                ]

        elif detect.isCentOS() or detect.isOracleLinux() or detect.isRHEL():
            commands += [
                '# package installation only',
                '/usr/bin/yum install *',
                '/usr/bin/yum-config-manager --add-repo *',
                '/usr/bin/yum-config-manager --enable *',
            ]

            if detect.isRHEL():
                commands += [
                    '# RHEL-specific',
                    '/sbin/subscription-manager repos --enable *',
                ]

            if not skip_key_mgmt:
                commands += [
                    '',
                    '# signing key setup',
                    '/usr/bin/rpm --import *',
                ]

        elif detect.isOpenSUSE() or detect.isSLES():
            commands += [
                '# package installation only',
                '/usr/bin/zypper --non-interactive install *',
                '/usr/bin/zypper --non-interactive addrepo *',
                '/usr/bin/zypper --non-interactive refresh *',
            ]

            if detect.isSLES():
                commands += [
                    '# package installation only',
                    '/usr/bin/SUSEConnect --product *',
                ]

            if not skip_key_mgmt:
                commands += [
                    '',
                    '# signing key setup',
                    '/bin/rpm --import *',
                ]

        elif detect.isArchLinux():
            commands += [
                '# package installation only',
                '/usr/bin/pacman *',
            ]

        elif detect.isGentoo():
            commands += [
                '# package installation only',
                '/usr/bin/emerge *',
            ]

        elif detect.isAlpineLinux():
            commands += [
                '# package installation only',
                '/sbin/apk add *',
                '/sbin/apk update',
                '/usr/bin/npm *',
                '/usr/bin/tee -a /etc/apk/repositories',
            ]

        elif detect.isMacOS():
            lines += [
                '# env whitelist',
                'Defaults  env_keep += "HOMEBREW_NO_GITHUB_API"',
            ]

            commands += [
                '# case for global brew install',
                '/usr/local/bin/brew install *',
                '/usr/local/bin/brew cask install *',
                '/usr/local/bin/brew search *',
                '/usr/local/bin/brew cask search *',
                '# note: security problem, if optional URL paremeter is used',
                '/usr/local/bin/brew tap homebrew/*',
                '/usr/local/bin/brew tap caskroom/*',
                '# lesser problem',
                '/usr/local/bin/brew unlink *',
            ]

            if not skip_key_mgmt:
                commands += [
                    '',
                    '# insecure, but required for local brew cask',
                    'SETENV: /usr/sbin/installer *',
                    'SETENV: /usr/sbin/pkgutil *',
                    'SETENV: /usr/libexec/PlistBuddy *',
                    'SETENV: /usr/bin/xargs -0 -- /bin/rm *',
                    'SETENV: /bin/mv *',
                    'SETENV: /bin/ln *',
                    'SETENV: /bin/chmod *',
                    'SETENV: /bin/chown *',
                    'SETENV: /bin/rm *',
                    'SETENV: /bin/rmdir *',
                ]

        else:
            self._errorExit('Unfortunately this OS is not fully supported yet')

        if ospath.exists('/bin/systemctl'):
            commands += [
                '',
                '# start of services',
                '/bin/systemctl start *',
            ]

        if ospath.exists('/sbin/rc-service'):
            commands += [
                '',
                '# start of services',
                '/sbin/rc-service * start',
            ]

        if detect.isLinux():
            commands += [
                '',
                '# allow access to Docker (may have no sense)',
                '/usr/bin/docker *',
            ]

        commands += [
            '',
            '# virtual env bootstrap',
            '/usr/bin/pip install *',
            '/usr/local/bin/pip install *',
        ]

        lines.append('')

        for c in commands:

            if c and c[0] != '#':
                lines.append('{0} ALL=(ALL) NOPASSWD: {1}'.format(entity, c))
            else:
                lines.append(c)

        lines.append('')
        lines = "\n".join(lines)
        print(lines)

    def builddep(self, deps):
        if not deps:
            res = self._builddep.available()
            print("\n".join(res))
            return

        deps_set = set(deps)
        tools = set(self._getKnownTools()) & deps_set
        self._config = {
            'tools': dict([(t, True) for t in tools])
        }
        self._initTools()

        env = self._env

        if self._detect.isExternalToolsSetup(env):
            self._executil.externalSetup(env, ['build-dep'] + deps)
        else:
            self._builddep.require(env, list(deps_set - tools))



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool
from .piptoolmixin import PipToolMixIn


class dockercomposeTool(PipToolMixIn, BuildTool, RunEnvTool):
    """Compose is a tool for defining and running multi-container Docker applications.

Home: https://docs.docker.com/compose/

Experimental support.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return ['docker-compose.yml', 'docker-compose.yaml']

    def getDeps(self):
        return ['pip', 'docker']

    def getOrder(self):
        return 20

    def _pipName(self):
        return 'docker-compose'

    def _installTool(self, env):
        self._requirePythonDev(env)
        super(dockercomposeTool, self)._installTool(env)

    def onBuild(self, config):
        cmd = [config['env']['dockercomposeBin'], 'build']
        self._executil.callMeaningful(cmd)

    def onPackage(self, config):
        cmd = [config['env']['dockercomposeBin'], 'bundle']
        self._executil.callMeaningful(cmd)

    def initEnv(self, env):
        super(dockercomposeTool, self).initEnv(env, 'docker-compose')



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .subtool import SubTool

__all__ = ['RuntimeTool']


class RuntimeTool(SubTool):
    __slots__ = ()
    DEFAULT_EXIT_TIMEOUT = 5000

    def __init__(self, name):
        super(RuntimeTool, self).__init__(name)

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([
            env[self._name + 'Bin'], svc['path']
        ] + args)

    def tuneDefaults(self, env):
        return {
            'minMemory': '1M',
            'connMemory': '1M',
            'scalable': False,
            'reloadable': False,
            'multiCore': False,
            'exitTimeoutMS': self.DEFAULT_EXIT_TIMEOUT,
            'maxRequestSize': '1M',
            'socketProtocol': 'custom',
        }

    def onStop(self, config, pid, tune):
        self._signalPID(pid, self._sigStop())

    def onReload(self, config, pid, tune):
        if tune['reloadable']:
            self._signalPID(pid, self._sigReload())
        else:
            self.onStop(config, pid, tune)

    def onPreConfigure(self, config, runtime_dir, svc, cfg_svc_tune):
        pass

    def _signalPID(self, pid, sig):
        self._os.kill(pid, sig)

    def _sigReload(self):
        return self._ext.signal.SIGHUP

    def _sigStop(self):
        return self._ext.signal.SIGTERM



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import wraps as _wraps


def simple_memo(f):
    @_wraps(f)
    def wrap():
        fdict = f.__dict__

        if 'cached_val' in fdict:
            return fdict['cached_val']

        val = f()
        fdict['cached_val'] = val
        return val

    return wrap


def complex_memo(f):
    f.cached = {}

    @_wraps(f)
    def wrap(*args, **kwargs):
        cached = f.cached
        key = str(args) + str(kwargs)
        if key in cached:
            return cached[key]

        val = f(*args, **kwargs)

        if val is not None:
            cached[key] = val

        return val

    return wrap



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool


class phpfpmTool(RuntimeTool):
    """PHP is a popular general-purpose scripting language that is especially suited to web development.

Home: http://php.net/

This tool provides PHP-FPM based website entry point support.
It means any PHP file in project can be executed with all consequences.

Use .tune.config for php-fpm.conf tuning - dicts of dicts representing ini file.
* 'global' -> FPM config
* 'pool' -> pool config

Use .tune.cid for default algorithm tuning
* 'extension' = [] - list of additional extensions to load
* 'zend_extension' - [] - list of additional zend extensions to load

Use .tune.phpini for php.ini tuning - dict representing ini file options.

Note: system php.ini is ignored, but all extensions are automatically loaded from system scan dir.
Note: file upload is OFF by default.
"""
    __slots__ = ()

    def getDeps(self):
        return ['php']

    def tuneDefaults(self, env):
        return {
            'minMemory': '1M',
            'connMemory': '16M',
            'debugConnOverhead': '16M',
            'socketTypes': ['unix', 'tcp'],
            'socketType': 'unix',
            'socketProtocol': 'fcgi',
            'scalable': True,
            'reloadable': False,
            'multiCore': True,
            'maxInstances': 2,
            'maxRequestSize': '1M',
        }

    def _installTool(self, env):
        self._phputil.installExtensions(env, 'opcache', True)

        php_ver = env['phpVer']

        if not env['phpSourceBuild']:
            self._installBinaries(env)
            return

    def _installBinaries(self, env):
        ver = env['phpVer']
        base_pkg = self._phputil.basePackage(ver)

        detect = self._detect

        if detect.isDebian() or detect.isUbuntu():
            self._install.deb('{0}-fpm'.format(base_pkg))

        elif detect.isSCLSupported():
            if self._phputil.isIUSVer(ver):
                self._install.yum('{0}-fpm'.format(base_pkg))
            else:
                self._install.yum('{0}-php-fpm'.format(base_pkg))

        else:
            fpm_pkg = '{0}-fpm'.format(base_pkg)
            # self._install.brew(nothing)
            self._install.zypper(fpm_pkg)
            self._install.yum(fpm_pkg)
            self._install.pacman(fpm_pkg)
            self._install.apk(fpm_pkg)

    def envNames(self):
        return ['phpfpmBin', 'phpfpmVer', 'phpfpmErrorLog']

    def _setFPMBin(self, env, phpfpm_bin):
        os = self._os
        ospath = self._ospath
        pathutil = self._pathutil

        if 'phpFakeBinDir' in env:
            php_bin_dir = env['phpFakeBinDir']
            orig_bin = phpfpm_bin
            phpfpm_bin = ospath.join(php_bin_dir, 'php-fpm')

            if ospath.exists(php_bin_dir) and ospath.islink(phpfpm_bin) and os.readlink(phpfpm_bin) == orig_bin:
                pass
            else:
                pathutil.rmTree(phpfpm_bin)
                os.symlink(orig_bin, phpfpm_bin)
        else:
            pathutil.addBinPath(ospath.dirname(phpfpm_bin), True)

        env['phpfpmBin'] = phpfpm_bin
        self._have_tool = True

    def initEnv(self, env):
        ospath = self._ospath
        detect = self._detect
        phputil = self._phputil

        php_ver = env['phpVer']
        phpfpm_ver = env.setdefault('phpfpmVer', php_ver)

        if php_ver != phpfpm_ver:
            self._errorExit(
                'PHP mismatch FPM version {0} != {1}'.format(php_ver, phpfpm_ver))
        #---
        if self._detect.isMacOS():
            # TODO: find solution
            error_log = '/dev/null'
        else:
            error_log = '/proc/self/fd/2'

        env.setdefault('phpfpmErrorLog', error_log)

        #---
        if 'phpBin' not in env:
            return

        php_bin = env['phpBin']

        if env['phpSourceBuild'] or detect.isSCLSupported():
            phpfpm_bin = ospath.join(php_bin, '..', '..', 'sbin', 'php-fpm')
            phpfpm_bin = ospath.realpath(phpfpm_bin)

            if ospath.exists(phpfpm_bin):
                self._setFPMBin(env, phpfpm_bin)
                return
        else:
            bin_name = 'php-fpm'

            if detect.isDebian() or detect.isUbuntu():
                bin_name = 'php-fpm' + php_ver

            elif detect.isArchLinux():
                if phputil.isArchLatest(php_ver):
                    bin_name = 'php-fpm'
                else:
                    bin_name = 'php{0}-fpm'.format(php_ver.replace('.', ''))

            elif detect.isAlpineLinux():
                if phputil.isAlpineSplit():
                    bin_name = 'php-fpm' + php_ver[0]
                else:
                    bin_name = 'php-fpm'

            #--
            phpfpm_bin = ospath.realpath(
                ospath.join(php_bin, '..', '..', 'sbin',
                            '{0}*'.format(bin_name))
            )
            phpfpm_bin = self._ext.glob.glob(phpfpm_bin)

            if phpfpm_bin:
                self._setFPMBin(env, phpfpm_bin[0])
                return

            # fallback, find any
            #---
            self._pathutil.addBinPath('/usr/sbin')
            phpfpm_bin = self._pathutil.which(bin_name)

            if phpfpm_bin:
                self._setFPMBin(env, phpfpm_bin)
                return

    def onPreConfigure(self, config, runtime_dir, svc, cfg_svc_tune):
        ospath = self._ospath
        configutil = self._configutil
        env = config['env']
        deploy = config['deploy']
        name_id = '{0}-{1}'.format(svc['name'], svc['instanceId'])

        svc_tune = svc['tune']
        max_clients = svc_tune['maxConnections']

        # conf location
        fpm_conf = "phpfpm-{0}.conf".format(name_id)
        fpm_conf = ospath.join(runtime_dir, fpm_conf)

        php_conf = "php-{0}.ini".format(name_id)
        php_conf = ospath.join(runtime_dir, php_conf)

        # Save location to deployment config
        cfg_svc_tune.update({
            'fpmConf': fpm_conf,
            'phpConf': php_conf,
        })

        #
        log_level = 'error'
        error_log = env['phpfpmErrorLog']
        display_errors = 'Off'
        error_reporting = 'E_ALL & ~E_NOTICE & ~E_STRICT & ~E_DEPRECATED'

        if config['env']['type'] != 'prod':
            log_level = 'notice'
            display_errors = 'On'
            error_reporting = 'E_ALL | E_STRICT'

        #
        fpm_ini = svc_tune.get('config', {})
        fpm_ini = self._ext.copy.deepcopy(fpm_ini)

        #
        cid_tune = svc_tune.get('cid', {})

        #
        global_ini = fpm_ini.setdefault('global', {})
        global_ini.setdefault('error_log', error_log)
        global_ini.setdefault('log_level', log_level)
        global_ini.setdefault(
            'syslog.ident', configutil.syslogTag(env, name_id))
        global_ini['daemonize'] = 'no'
        global_ini['rlimit_files'] = svc_tune['maxFD']

        if ospath.exists('/bin/systemctl'):
            global_ini['systemd_interval'] = '0'

        #
        pool_ini = fpm_ini.setdefault('pool', {})

        socket_type = svc_tune['socketType']

        if socket_type == 'unix':
            pool_ini['listen'] = svc_tune['socketPath']
        elif socket_type == 'tcp':
            pool_ini['listen'] = '{0}:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])
        else:
            self._errorExit(
                'Unsupported socket type "{0}" for "{1}"'.format(socket_type, name_id))

        pool_ini.setdefault('listen.backlog', -1)
        pool_ini['user'] = deploy['user']
        pool_ini['group'] = deploy['group']
        pool_ini.setdefault('pm', 'static')
        pool_ini['pm.max_children'] = max_clients
        pool_ini['chdir'] = config['wcDir']
        pool_ini.setdefault('clear_env', 'yes')
        pool_ini.setdefault('catch_workers_output', 'on')

        self._pathutil.writeIni(fpm_conf, fpm_ini)

        #
        tmp_dir = ospath.join(deploy['tmpDir'], 'php')
        self._pathutil.mkDir(tmp_dir)

        upload_dir = ospath.join(deploy['tmpDir'], 'phpupload')
        self._pathutil.mkDir(upload_dir)

        #
        php_ini = svc_tune.get('phpini', {})
        php_ini = php_ini.copy()

        php_ini.setdefault('sys_temp_dir', tmp_dir)
        php_ini.setdefault(
            'memory_limit', configutil.parseMemory(svc_tune['connMemory']))
        php_ini.setdefault('expose_php', 'Off')
        php_ini.setdefault('zend.multibyte', 'On')
        php_ini.setdefault('zend.script_encoding', 'UTF-8')
        php_ini.setdefault('default_charset', 'UTF-8')
        php_ini.setdefault('register_globals', 'Off')
        php_ini.setdefault('include_path', '.')
        php_ini.setdefault('error_log', error_log)
        php_ini.setdefault('display_errors', display_errors)
        php_ini.setdefault('display_startup_errors', display_errors)
        php_ini.setdefault('error_reporting', error_reporting)
        php_ini.setdefault('file_uploads', 'On')
        php_ini.setdefault('upload_tmp_dir', upload_dir)
        request_size_limit = configutil.parseMemory(
            svc_tune['maxRequestSize'])
        php_ini['upload_max_filesize'] = request_size_limit
        php_ini['post_max_size'] = request_size_limit

        if config['env']['type'] != 'dev':
            php_ini.setdefault('open_basedir', config['deployDir'])

        # note: basic extensions are loaded from system (PHP_INI_SCAN_DIR)
        for k in ('extension', 'zend_extension'):
            if k in cid_tune:
                php_ini[k] = cid_tune

        self._pathutil.writeIni(php_conf, {'php': php_ini})

        # Validate
        self._executil.callExternal([
            env['phpfpmBin'],
            '-c', php_conf,
            '--fpm-config', fpm_conf,
            '--test',
        ])

    def onRun(self, config, svc, args):
        env = config['env']
        svc_tune = svc['tune']
        self._executil.callInteractive([
            env['phpfpmBin'],
            '-c', svc_tune['phpConf'],
            '--fpm-config', svc_tune['fpmConf'],
            '--nodaemonize',
        ] + args)
